#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "l2.h"
#include "getsym.h"
#include "dts.h"

#define NELEMS(a) (sizeof(a) / sizeof(a[0]))

#define MAXABBLEN 11            /* the maximum length in abb */

static const struct {const char *str; size_t len;} abb[] =
{
  {"A.D.", 4}, {"A.M.", 4}, {"Adm.", 4}, {"Adr.tel.", 8}, {"Al.", 3},
  {"Am.Pd.", 6}, {"Am.Pn.", 6}, {"Amer.P\263d.", 9},
  {"Amer.P\263n.", 9}, {"Archip.", 7}, {"Az.", 3}, {"Bl.", 3},
  {"Bryg.", 5}, {"C.O.", 4}, {"Dw.", 3}, {"Dyr.", 4}, {"Dz.U.", 5},
  {"Fot.", 4}, {"Fr.", 3}, {"G.", 2}, {"I.I.", 4}, {"ILek.", 5},
  {"ILot.", 5}, {"IPGum.", 6}, {"Inst.", 5}, {"J.", 2}, {"J.M.", 4},
  {"J.O.", 4}, {"J.P.", 4}, {"J.PP.", 5}, {"Jez.", 4}, {"Kan.", 4},
  {"Kier.", 5}, {"M.", 2}, {"Min.", 4}, {"N.", 2}, {"NB.", 3},
  {"Niz.", 4}, {"Np.", 3}, {"Nr rej.", 7}, {"O.", 2}, {"Ob.", 3},
  {"Oc.", 3}, {"Os.", 3}, {"O\266r.", 4}, {"P.", 2}, {"P.M.", 4},
  {"P.T.", 4}, {"PP.", 3}, {"PTAr.", 5}, {"Pd.", 3}, {"Pl.", 3},
  {"Prez.", 5}, {"P\263w.", 4}, {"P\363\263.", 4}, {"Rp.", 3},
  {"Rz.P.", 5}, {"Rzpl.", 5}, {"S.A.", 4}, {"S.T.", 4}, {"St.", 3},
  {"St.Zjedn.", 9}, {"Stow.", 5}, {"Szt.G\263.", 7}, {"T.T.", 4},
  {"Tow.", 4}, {"W.", 2}, {"W.Bryt.", 7}, {"Woj.", 4}, {"Wy\277.", 4},
  {"Zarz.G\263.", 8}, {"Zat.", 4}, {"Zw.", 3}, {"Zw.Radz.", 8},
  {"a.C.", 4}, {"a.C.n.", 6}, {"a.Chr.", 6}, {"a.Chr.n.", 8},
  {"a.a.C.", 6}, {"a.c.", 4}, {"a.h.l.", 6}, {"a.m.", 4}, {"a.p.C.", 6},
  {"ab init.", 8}, {"abp.", 4}, {"abs.", 4}, {"ad lib.", 7},
  {"ad loc.", 7}, {"adapt.", 6}, {"adj.", 4}, {"adm.", 4},
  {"adr.telegr.", 11}, {"adw.", 4}, {"agd.", 4}, {"al.", 3}, {"am.", 3},
  {"amb.", 4}, {"amer.", 5}, {"ang.", 4}, {"arch.", 5}, {"archip.", 7},
  {"arcybp.", 7}, {"arcyks.", 7}, {"art.", 4}, {"art.mal.", 8},
  {"art.rze\274.", 9}, {"aud.", 4}, {"b.", 2}, {"b.z.", 4},
  {"ba\263k.", 5}, {"ba\263t.", 5}, {"biul.", 5}, {"bl.", 3},
  {"bm.", 3}, {"bosm.", 5}, {"bp.", 3}, {"br.", 3}, {"bryg.", 5},
  {"bryt.", 5}, {"b\263.", 3}, {"c.", 2}, {"c.at.", 5}, {"c.cz.", 5},
  {"c.o.", 4}, {"c.par.", 6}, {"c.w\263.", 5}, {"cf.", 3}, {"chor.", 5},
  {"ci\352\277.", 5}, {"cz.", 3}, {"czyt.", 5}, {"cz\263.", 4},
  {"del.", 4}, {"dew.", 4}, {"dn.", 3}, {"doc.", 4}, {"dod.", 4},
  {"dol.", 4}, {"dop\263.", 5}, {"dos\263.", 5}, {"dot.", 4},
  {"dr h.c.", 7}, {"dr hab.", 7}, {"dr med.", 7}, {"dr.", 3},
  {"ds.", 3}, {"dwumies.", 8}, {"dwustr.", 7}, {"dwuszp.", 7},
  {"dwutyg.", 7}, {"dypl.", 5}, {"dyr.", 4}, {"dz.", 3}, {"dziek.", 6},
  {"dzien.", 6}, {"d\263.", 3}, {"d\263.geogr.", 9}, {"e.c.", 4},
  {"egz.", 4}, {"ekg.", 4}, {"ekw.", 4}, {"em.", 3}, {"err.", 4},
  {"ew.", 3}, {"ewent.", 6}, {"f.d\263.", 5}, {"f.kr.", 5},
  {"f.szt.", 6}, {"f.\266r.", 5}, {"fel.", 4}, {"fl.", 3}, {"flor.", 5},
  {"fot.", 4}, {"fotom.", 6}, {"fotorep.", 8}, {"fr.", 3},
  {"fragm.", 6}, {"g.", 2}, {"gat.", 4}, {"gen.bryg.", 9},
  {"gen.dyw.", 8}, {"godz.", 5}, {"gosp.", 5}, {"grub.", 5},
  {"g\263.", 3}, {"h.a.", 4}, {"h.c.", 4}, {"hon.", 4}, {"hr.", 3},
  {"i.e.", 4}, {"i.m.", 4}, {"ib.", 3}, {"ibid.", 5}, {"id.", 3},
  {"ilustr.", 7}, {"im.w\263.", 6}, {"in.", 3}, {"in\277.", 4},
  {"in\277.agr.", 8}, {"in\277.arch.", 9}, {"in\277.chem.", 9},
  {"in\277.elektr.", 11}, {"in\277.g\363r.", 8}, {"in\277.hut.", 8},
  {"in\277.in\277.", 8}, {"in\277.le\266n.", 9}, {"in\277.mech.", 9},
  {"j.", 2}, {"j.em.", 5}, {"j.es.", 5}, {"j.m.", 4}, {"jedn.", 5},
  {"jez.", 4}, {"jun.", 4}, {"j\352z.", 4}, {"j\352z.oryg.", 9},
  {"k.", 2}, {"k.c.", 4}, {"k.k.", 4}, {"k.o.", 4}, {"k.p.a.", 6},
  {"k.p.c.", 6}, {"k.p.k.", 6}, {"k.r.", 4}, {"k.r.i o.", 8},
  {"k.r.o.", 6}, {"k.tyt.", 6}, {"k.z.", 4}, {"kad.", 4}, {"kal.", 4},
  {"kan.", 4}, {"kanc.", 5}, {"kand.n.", 7}, {"kard.", 5},
  {"kartogr.", 8}, {"kc.", 3}, {"kk.", 3}, {"kl.", 3}, {"kol.red.", 8},
  {"kom.", 4}, {"kom.red.", 8}, {"koment.", 7}, {"koresp.", 7},
  {"ko\266c.", 5}, {"kpc.", 4}, {"kpr.", 4}, {"kpt.", 4}, {"kr.", 3},
  {"ks.", 3}, {"ksi\352g.", 6}, {"kw.", 3}, {"kz.", 3}, {"l.", 2},
  {"l.a.", 4}, {"l.at.", 5}, {"l.c.", 4}, {"l.dz.", 5}, {"l.l.", 4},
  {"l.ub.", 5}, {"litogr.", 7}, {"lotn.", 5}, {"ludn.", 5},
  {"m kw.", 5}, {"m.", 2}, {"m.at.", 5}, {"m.b.", 4}, {"m.cz.", 5},
  {"m.in.", 5}, {"m.st.", 5}, {"m.woj.", 6}, {"maks.", 5}, {"mal.", 4},
  {"margr.", 6}, {"med.", 4}, {"mgr.", 4}, {"mies.", 5}, {"mieszk.", 7},
  {"mjr.", 4}, {"m\263.", 3}, {"n.", 2}, {"n.e.", 4}, {"n.p.m.", 6},
  {"nak\263.", 5}, {"nb.", 3}, {"niz.", 4}, {"np.", 3}, {"nt.", 3},
  {"ob.", 3}, {"obj.", 4}, {"obr.", 4}, {"obwol.", 6}, {"odb.", 4},
  {"odc.", 4}, {"oddz.", 5}, {"ok.", 3}, {"okr.", 4}, {"op.", 3},
  {"opr.", 4}, {"ork.", 4}, {"oryg.", 5}, {"osk.", 4}, {"o\266r.", 4},
  {"p.", 2}, {"p.Ch.n.", 7}, {"p.Chr.n.", 8}, {"p.cz.", 5},
  {"p.k.s.", 6}, {"p.m.", 4}, {"p.n.", 4}, {"p.n.e.", 6}, {"p.o.", 4},
  {"p.p.m.", 6}, {"pchor.", 6}, {"pd.", 3}, {"pd.-wsch.", 9},
  {"pd.-zach.", 9}, {"pks.", 4}, {"pkt.", 4}, {"pl.", 3}, {"plut.", 5},
  {"pn.", 3}, {"pn.-wsch.", 9}, {"pn.-zach.", 9}, {"pok.", 4},
  {"polit.", 6}, {"pos.", 4}, {"pow.", 4}, {"poz.", 4}, {"po\263.", 4},
  {"po\263ud.", 6}, {"pp.", 3}, {"ppanc.", 6}, {"ppor.", 5},
  {"pp\263k.", 5}, {"pr.", 3}, {"prez.", 5}, {"prob.", 5}, {"prof.", 5},
  {"proj.", 5}, {"prosp.", 6}, {"przedst.", 8}, {"przet\263.", 7},
  {"przew.", 6}, {"prze\263.", 6}, {"przyg.", 6}, {"przyw.", 6},
  {"ps.", 3}, {"pt.", 3}, {"p\263.", 3}, {"p\263d.", 4}, {"p\263k.", 4},
  {"p\263n.", 4}, {"p\263p\263t.", 6}, {"p\263t.", 4}, {"p\263w.", 4},
  {"p\363\263sk.", 6}, {"p\363\263w.", 5}, {"q.v.", 4}, {"r.", 2},
  {"r.szk.", 6}, {"r.ub.", 5}, {"radz.", 5}, {"rb.", 3}, {"rec.", 4},
  {"red.nacz.", 9}, {"red.nauk.", 9}, {"ref.", 4}, {"re\277.", 4},
  {"ros.", 4}, {"rozdz.", 6}, {"rtg.", 4}, {"rtm.", 4}, {"rubr.", 5},
  {"ryc.", 4}, {"rz.", 3}, {"s.", 2}, {"s.c.", 4}, {"s.o.s.", 6},
  {"san.-epid.", 10}, {"sekr.", 5}, {"sier\277.", 6}, {"sk.", 3},
  {"sp.z o.o.", 9}, {"sp.z o.p.", 9}, {"spo\263.", 5}, {"ss.", 3},
  {"st.", 3}, {"st.mar.", 7}, {"st.ogn.", 7}, {"st.sier\277.", 9},
  {"st.strz.", 8}, {"st.szer.", 8}, {"stow.", 5}, {"str.", 4},
  {"strz.", 5}, {"sygn.", 5}, {"szer.", 5}, {"szer.geogr.", 11},
  {"sze\266c.", 6}, {"szt.", 4}, {"t.", 2}, {"t.m.", 4}, {"tab.", 4},
  {"tabl.", 5}, {"techn.", 6}, {"tel.", 4}, {"telegr.", 7},
  {"temp.", 5}, {"tj.", 3}, {"tow.", 4}, {"tw.", 3}, {"tw.szt.", 7},
  {"tys.", 4}, {"tyt.", 4}, {"tzn.", 4}, {"tzw.", 4}, {"ub.", 3},
  {"ub.m.", 5}, {"ub.r.", 5}, {"uniw.", 5}, {"ur.", 3}, {"usg.", 4},
  {"v.", 2}, {"w st.sp.", 8}, {"w.", 2}, {"w.c.", 4}, {"wach.", 5},
  {"wewn.", 5}, {"wf.", 3}, {"wicemin.", 8}, {"woj.", 4}, {"wsch.", 5},
  {"ww.", 3}, {"wym.", 4}, {"wys.", 4}, {"wz.", 3}, {"z o.o.", 6},
  {"z.", 2}, {"zach.", 5}, {"zam.", 4}, {"zat.", 4}, {"za\263.", 4},
  {"zewn.", 5}, {"zm.", 3}, {"zob.", 4}, {"zw.", 3}, {"zw.zaw.", 7},
  {"zw\263.", 4}, {"z\263 dew.", 7}, {"\246w.", 3}, {"\263ac.", 4},
  {"\266.", 2}, {"\266p.", 3}, {"\266r.", 3}, {"\266rodk.", 6},
  {"\266w.", 3}, {NULL, 0}
};

static int isabb(struct parse *p)
{
  char s[MAXABBLEN + 1];
  int i = p->n, n1 = i - 1, h;
  int avail = MAXABBLEN, request;
  size_t ls;

  while (i-- > 0)
  {
    request = strlen(p->run[i].val);
    if ((p->run[i].type == WORD || p->run[i].type == NUMBER)
        && i < n1 && p->run[i].type == p->run[i + 1].type)
      request++;
    if (request > avail)
    {
      i++;
      break;
    }
    avail -= request;
  }
  *s = '\0';
  while (i < p->n)
  {
    strcat(s, p->run[i].val);
    if ((p->run[i].type == WORD || p->run[i].type == NUMBER)
        && i < n1 && p->run[i].type == p->run[i + 1].type)
      strcat(s, " ");
    i++;
  }
  ls = strlen(s);
  for (i = 0; abb[i].str != NULL; i++)
    if (ls >= abb[i].len)
    {
      h = ls - abb[i].len;
      if (strcmp(s + h, abb[i].str) == 0)
        if (h == 0 || !l2_isalpha(*(s + h - 1)))
          return 1;
    }
  return 0;
}

static void getlex(struct parse *p)
{
  struct symbol *s;

  assert(!((p->first == 0 && p->last == NELEMS(p->que) - 1) ||
           p->first == p->last + 1));
  if (p->last == NELEMS(p->que) - 1 || p->last == -1)
  {
    p->last = 0;
    if (p->first == -1)
      p->first = 0;
  }
  else
    p->last++;
  s = &p->que[p->last];
  s->type = getsym(p->inputf, s->val, NELEMS(s->val), &p->line,
                   &p->status);
}

static void move(struct parse *p)
{
  struct symbol *s = &p->que[p->first], *t = &p->run[p->n];

  assert(p->first != -1);
  assert(p->n < NELEMS(p->run));
  t->type = s->type;
  strcpy(t->val, s->val);
  p->n++;
  if (p->first == p->last)
    p->first = p->last = -1;
  else if (p->first == NELEMS(p->que) - 1)
    p->first = 0;
  else
    p->first++;
}

static int isZ(int t)
{
  switch (t)
  {
    case FULL_STOP: case SEMICOLON: case COMMA: case COLON: case DOTS:
    case QUESTION_MARK: case EXCLAMATION_MARK: case RIGHT_ROUND_BRACKET:
    case RIGHT_SQUARE_BRACKET: case RIGHT_CURLY_BRACKET:
    case RIGHT_ANGLE_BRACKET: case RIGHT_QUOTATION_MARK:
    case RIGHT_ANGLE_QUOTATION_MARK:
    case RIGHT_DEFINITION_QUOTATION_MARK: case HYPHEN:
      return 1;
  }
  return 0;
}

static void get_next_run(struct parse *p)
{
  int i, x, y, z;

  p->type = 0;
  if (p->first == -1)
    getlex(p);
  if (p->que[p->first].type == 0)
    return;
  p->n = 0;
  p->type = 2;
  if (isZ(p->que[p->first].type))
    for (;;)
    {
      move(p);
      if (p->first == -1)
        getlex(p);
      if (!isZ(p->que[p->first].type))
        return;
    }
  p->type = 1;
  for (;;)
  {
    move(p);                    /* step 1 */
    if (p->first == -1)
      getlex(p);
    y = p->run[p->n - 1].type;
    if (y != FULL_STOP && y != QUESTION_MARK && y != EXCLAMATION_MARK &&
        y != DOTS)
    {
      if (p->que[p->first].type == 0)
      {
        p->type = 2;
        return;
      }
      else
        continue;
    }
    if (p->que[p->first].type == 0)
      return;                   /* step 2 */
    if (isZ(p->que[p->first].type))
      continue;                 /* step 3 */
    if ((i = p->first) != -1)
      for (;;)
      {
        z = p->que[i].type;
        if (z == WORD || z == NUMBER)
          break;
        else
        {
          if (i == p->last)
          {
            i = -1;
            break;
          }
          if (i == NELEMS(p->que) - 1)
            i = 0;
          else
            i++;
        }
      }
    if (i == -1)
      for (;;)
      {
        getlex(p);
        i = p->last;
        z = p->que[i].type;
        if (z == 0)
          return;               /* step 4 */
        if (z == WORD || z == NUMBER)
          break;
      }
    if (l2_islower(p->que[i].val[0]))
    {
      while (p->first != i)
        move(p);
      continue;                 /* step 5 */
    }
    if (y != FULL_STOP)
      return;                   /* step 6 */
    assert(p->n > 1);
    x = p->run[p->n - 2].type;
    if (x != WORD && x != NUMBER)
      return;                   /* step 7 */
    if (x == NUMBER)
    {
      if (z == WORD)
        return;                 /* step 8a */
      if (z == NUMBER)
        continue;               /* step 8b */
    }
    if (isabb(p))
      continue;                 /* step 8c */
    if (strlen(p->run[p->n - 2].val) == 1 &&
        l2_isupper(p->run[p->n - 2].val[0]))
    {
      if (z == WORD)
        continue;               /* step 8d */
      if (z == NUMBER)
        return;                 /* step 8e */
    }
    return;                     /* step 8f */
  }
}

int dts(struct parse *p)
{
  get_next_run(p);
  if (p->type == 1)
  {
    int i;
    for (i = 0; i < p->n; i++)
      if ((p->run[i].type == WORD && l2_isupper(p->run[i].val[0])) ||
          p->run[i].type == NUMBER)
        break;
    if (i >= p->n)
      p->type = 2;
  }
  return p->type;
}

#ifdef TEST_DTS

int main(void)
{
  struct parse p;
  int i;

  p.line = p.status = 0;
  p.first = p.last = -1;
  p.inputf = stdin;
  while (dts(&p) != 0)
    if (p.type == 1)
    {
      for (i = 0; i < p.n; i++)
        printf("%s\n", p.run[i].val);
      printf("\n");
    }
  assert(feof(p.inputf) && !ferror(p.inputf));
  return 0;
}

#endif /* TEST_DTS */
