#include <assert.h>
#include <stdio.h>
#include "l2.h"
#include "getsym.h"

static int pmtype1(int c)
{
  if (c == 0x2e)
    return FULL_STOP;
  else if (c == 0x2c)
    return COMMA;
  else if (c == 0x2d)
    return DASH;
  else if (c == 0x27)
    return RIGHT_DEFINITION_QUOTATION_MARK;
  else if (c == 0x3c)
    return LEFT_ANGLE_BRACKET;
  else if (c == 0x3e)
    return RIGHT_ANGLE_BRACKET;
  else
    return 0;
}

static int pmtype2(int c)
{
  if (c == 0x3b)
    return SEMICOLON;
  else if (c == 0x3a)
    return COLON;
  else if (c == 0x3f)
    return QUESTION_MARK;
  else if (c == 0x21)
    return EXCLAMATION_MARK;
  else if (c == 0x28)
    return LEFT_ROUND_BRACKET;
  else if (c == 0x29)
    return RIGHT_ROUND_BRACKET;
  else if (c == 0x5b)
    return LEFT_SQUARE_BRACKET;
  else if (c == 0x5d)
    return RIGHT_SQUARE_BRACKET;
  else if (c == 0x2f)
    return SIMPLE_BRACKET;
  else if (c == 0x7b)
    return LEFT_CURLY_BRACKET;
  else if (c == 0x7d)
    return RIGHT_CURLY_BRACKET;
  else if (c == 0x60)
    return LEFT_DEFINITION_QUOTATION_MARK;
  else if (c == 0x22)
    return TYPEWRITER_QUOTATION_MARK;
  else
    return 0;
}

static int gstype(int c)
{
  if (c == 0x3d)
    return EQUALS_SIGN;
  else if (c == 0x2b)
    return PLUS_SIGN;
  else if (c == 0x25)
    return PERCENT_SIGN;
  else if (c == 0xa7)
    return PARAGRAPH_SIGN;
  else if (c == 0x23)
    return NUMBER_SIGN;
  else if (c == 0x24)
    return DOLLAR_SIGN;
  else if (c == 0x26)
    return AMPERSAND;
  else if (c == 0x2a)
    return ASTERISK;
  else if (c == 0x40)
    return COMMERCIAL_AT;
  else if (c == 0x5c)
    return BACKSLASH;
  else if (c == 0x5e)
    return UP_ARROW;
  else if (c == 0x5f)
    return UNDERLINE;
  else if (c == 0x7c)
    return VERTICAL_LINE;
  else if (c == 0x7e)
    return TILDE;
  else
    return 0;
}

int getsym(FILE *f, char *val, int n, int *line, int *st)
{
  int c, i = 0;
  int sym;

  n--;

  assert(n >= 3);

  if (*st != 0)
  {
    *st = 0;
    val[0] = 0x2e;              /* full stop */
    val[1] = '\0';
    return FULL_STOP;
  }

  while (l2_isspace((c = fgetc(f))))
    if (c == 0x0a)              /* new line */
      (*line)++;

  if (l2_isalpha(c))
  {
    do
    {
      if (i < n)
        val[i++] = c;
      else
        return 0;
      c = fgetc(f);
    } while (l2_isalpha(c));
    if (c != EOF && ungetc(c, f) == EOF)
      return 0;
    val[i] = '\0';
    return WORD;
  }
  else if ((sym = pmtype1(c)) != 0)
  {
    val[i++] = c;
    c = fgetc(f);
    if (c != val[0])
    {
      if (c != EOF && ungetc(c, f) == EOF)
        return 0;
      val[i] = '\0';
      return sym == DASH ? HYPHEN : sym;
    }
    val[i++] = c;
    if (c != 0x2e)              /* not a full stop */
    {
      val[i] = '\0';
      if (sym == COMMA)
        return LEFT_QUOTATION_MARK;
      if (sym == RIGHT_DEFINITION_QUOTATION_MARK)
        return RIGHT_QUOTATION_MARK;
      if (sym == LEFT_ANGLE_BRACKET)
        return LEFT_ANGLE_QUOTATION_MARK;
      if (sym == RIGHT_ANGLE_BRACKET)
        return RIGHT_ANGLE_QUOTATION_MARK;
      return sym;
    }
    c = fgetc(f);
    if (c != 0x2e)              /* not a full stop */
    {
      if (c != EOF && ungetc(c, f) == EOF)
        return 0;
      *st = 1;
      val[--i] = '\0';
      return FULL_STOP;
    }
    else
    {
      val[i++] = c;
      val[i] = '\0';
      return DOTS;
    }
  }
  else if ((sym = pmtype2(c)) != 0)
  {
    val[i++] = c;
    val[i] = '\0';
    return sym;
  }
  else if (l2_isdigit(c))
  {
    do
    {
      if (i < n)
        val[i++] = c;
      else
        return 0;
      c = fgetc(f);
    } while (l2_isdigit(c));
    if (c != EOF && ungetc(c, f) == EOF)
      return 0;
    val[i] = '\0';
    return NUMBER;
  }
  else if ((sym = gstype(c)) != 0)
  {
    val[i++] = c;
    val[i] = '\0';
    return sym;
  }
  else if (c == EOF)
    return 0;
  else
  {
    val[i++] = c;
    val[i] = '\0';
    return UNKNOWN;
  }
}
